@@ -2,10 +2,11 @@
  * @package      hubzero-telequotad
  * @file         telequotad.c
  * @author       Rick Kennell <kennell@purdue.edu>
- * @copyright    Copyright (c) 2005-2012 HUBzero Foundation, LLC.
+ * @author       Nicholas J. Kisseberth <nkissebe@purdue.edu>
+ * @copyright    Copyright (c) 2005-2013 HUBzero Foundation, LLC.
  * @license      http://www.gnu.org/licenses/lgpl-3.0.html LGPLv3
  *
- * Copyright (c) 2005-2012 HUBzero Foundation, LLC.
+ * Copyright (c) 2005-2013 HUBzero Foundation, LLC.
  *
  * This file is part of: The HUBzero(R) Platform for Scientific Collaboration
  *
@@ -88,10 +89,11 @@ static int64_t default_hardinode = 0;
 static int64_t default_softinode = 0;
 static char purgecmd[MAXPATHLEN] = "";
 static int listen_port = 0;
-static int sock = 0;
+static int sock4 = 0;
+static int sock6 = 0;
 
 struct allowed_addr {
-  struct in_addr addr;
+  struct sockaddr_storage addr;
   struct allowed_addr *next;
 };
 
@@ -355,26 +357,52 @@ int eatnum(char **pp, int64_t *result)
   return 1;
 }
 
-void allow(struct in_addr addr)
+void allow(struct sockaddr *addr)
 {
+  char straddr[INET6_ADDRSTRLEN];
+  struct sockaddr_in6 *sin6;
+  struct sockaddr_in *sin;
+  int len = 0;
+
+  sin = (struct sockaddr_in *) addr; 
+  sin6 = (struct sockaddr_in6 *) addr;
+    
+  if (addr->sa_family == AF_INET) {
+    inet_ntop(AF_INET, &sin->sin_addr, straddr, sizeof(straddr));
+    len = sizeof(struct sockaddr_in);
+  }
+  else if (addr->sa_family == AF_INET6) {
+    inet_ntop(AF_INET6, &sin6->sin6_addr, straddr, sizeof(straddr));
+    len = sizeof(struct sockaddr_in6);
+  }
+  else {
+      return;
+  }
+
   struct allowed_addr *p;
   for(p=allowed_list; p; p=p->next) {
-    if (memcmp(&p->addr,&addr,sizeof(addr)) == 0) {
-      logmsg(LOG_INFO,"Duplicate allowed address: %s\n", inet_ntoa(addr));
+    if (memcmp(&p->addr,addr,len) == 0) {
+      logmsg(LOG_INFO,"Duplicate allowed address: %s\n", straddr);
       return;
     }
   }
 
-  logmsg(LOG_INFO,"Allowing %s\n", inet_ntoa(addr));
+  logmsg(LOG_INFO,"Allowing %s\n", straddr);
   struct allowed_addr *next = calloc(sizeof(struct allowed_addr),1);
   next->next = allowed_list;
-  memcpy(&next->addr, &addr, sizeof(addr));
+  memcpy(&next->addr, addr, len);
 
   allowed_list = next;
 }
 
 void parse_allow(char **pp, int lineno)
 {
+  struct addrinfo hints;
+  struct addrinfo *hostaddr = NULL;
+  struct addrinfo *tmpaddr = NULL;
+  struct in6_addr server6addr;
+  struct in_addr server4addr;
+  int rv = 0;
   char name[100];
   eatword(pp,name,sizeof(name));
   if (name == NULL) {
@@ -382,32 +410,45 @@ void parse_allow(char **pp, int lineno)
     exit(1);
   }
 
-  struct in_addr addr;
+  memset(&hints, 0, sizeof(hints));
 
-  if (inet_aton(name, &addr)) {
-    allow(addr);
-    return;
+  hints.ai_family = PF_UNSPEC;
+  hints.ai_socktype = SOCK_STREAM;
+  hints.ai_flags = AI_NUMERICSERV;
+  
+  rv = inet_pton(AF_INET, name, &server6addr);
+
+  if (rv == 1) {
+    hints.ai_family = AF_INET;
+    hints.ai_flags |= AI_NUMERICHOST;
+  }
+  else {
+    rv = inet_pton(AF_INET6, name, &server4addr);
+
+    if (rv == 1) {
+      hints.ai_family = AF_INET6;
+      hints.ai_flags |= AI_NUMERICHOST;
+    }
   }
 
-  struct hostent *h = gethostbyname(name);
-  if (h == NULL) {
-    logmsg(LOG_ERR,"Unknown host '%s' in line %d\n", name, lineno);
-    exit(1);
-  }
-  if (h->h_length != sizeof(addr)) {
-    logmsg(LOG_ERR,"Line %d yeilds an IPv6 address.\n", lineno);
+  rv = getaddrinfo(name, NULL, &hints, &hostaddr);
+
+  if (rv != 0) {
+    logmsg(LOG_ERR, "Unable to get host lookup information for %s", name);
     exit(1);
   }
 
-  char **p;
-  for(p = h->h_addr_list; *p != NULL; p++) {
-    memcpy(&addr, *p, sizeof(addr));
-    allow(addr);
+  for(tmpaddr = hostaddr; tmpaddr; tmpaddr = tmpaddr->ai_next) {
+    if ( (tmpaddr->ai_family != AF_INET6) && (tmpaddr->ai_family != AF_INET) ) {
+      continue;
+    }
+ 
+    allow(tmpaddr->ai_addr);
   }
 
+  freeaddrinfo(hostaddr);
 }
 
-
 void parse_default(char **pp, int lineno)
 {
   if (eatsym(pp,"hardspace")) {
@@ -482,63 +523,160 @@ void print_config()
 
 void setup_listener()
 {
-  sock = socket(PF_INET, SOCK_STREAM, IPPROTO_IP);
-  if (sock < 0) {
+  sock6 = socket(PF_INET6, SOCK_STREAM, IPPROTO_IP);
+
+  sock4 = socket(PF_INET, SOCK_STREAM, IPPROTO_IP);
+
+  if (sock4 < 0 && sock6 < 0) {
     posixerror("socket");
-    exit(1);
+    exit(1);    
   }
 
-  int value = 1;
-  if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &value, sizeof(value)) != 0) {
-    posixerror("setsockopt SO_REUSEADDR");
-    exit(1);
+  if (sock4 >= 0) {
+    int value = 1;
+    if (setsockopt(sock4, SOL_SOCKET, SO_REUSEADDR, &value, sizeof(value)) != 0) {
+      posixerror("setsockopt SO_REUSEADDR");
+      exit(1);
+    }
+
+    struct linger lin = { 0, 0 };
+    if (setsockopt(sock4, SOL_SOCKET, SO_LINGER, &lin, sizeof(lin)) != 0) {
+      posixerror("setsockopt SO_LINGER");
+      exit(1);
+    }
+
+    struct sockaddr_in addr;
+    addr.sin_port = htons(listen_port);
+    addr.sin_addr.s_addr = INADDR_ANY;
+    if (bind(sock4, &addr, sizeof(addr))) {
+      posixerror("bind");
+      exit(1);
+    }
+
+    if (listen(sock4,20) != 0) {
+      posixerror("listen");
+      exit(1);
+    }
   }
 
-  struct linger lin = { 0, 0 };
-  if (setsockopt(sock, SOL_SOCKET, SO_LINGER, &lin, sizeof(lin)) != 0) {
-    posixerror("setsockopt SO_LINGER");
-    exit(1);
-  }
+  if (sock6 >= 0) {
+    int value = 1;
+    if (setsockopt(sock6, SOL_SOCKET, SO_REUSEADDR, &value, sizeof(value)) != 0) {
+      posixerror("setsockopt SO_REUSEADDR");
+      exit(1);
+    }
 
+    struct linger lin = { 0, 0 };
+    if (setsockopt(sock6, SOL_SOCKET, SO_LINGER, &lin, sizeof(lin)) != 0) {
+      posixerror("setsockopt SO_LINGER");
+      exit(1);
+    }
 
-  struct sockaddr_in addr;
-  addr.sin_port = htons(listen_port);
-  addr.sin_addr.s_addr = 0;
-  if (bind(sock, &addr, sizeof(addr))) {
-    posixerror("bind");
-    exit(1);
-  }
+    value = 1;
+    if (setsockopt(sock6, IPPROTO_IPV6, IPV6_V6ONLY, &value, sizeof(value)) != 0) {
+      posixerror("setsockopt IPV6_V6ONLY");
+      exit(1);
+    }
 
-  if (listen(sock,20) != 0) {
-    posixerror("listen");
-    exit(1);
+    struct sockaddr_in6 addr;
+    addr.sin6_port = htons(listen_port);
+    addr.sin6_addr = in6addr_any;
+    if (bind(sock6, &addr, sizeof(addr))) {
+      posixerror("bind");
+      exit(1);
+    }
+
+    if (listen(sock6,20) != 0) {
+      posixerror("listen");
+      exit(1);
+    }
   }
 }
 
 
 void handle_connections()
 {
+  fd_set active_fd_set, read_fd_set;
+
+  FD_ZERO(&active_fd_set);
+
+  if (sock4 >= 0) {
+    FD_SET(sock4, &active_fd_set);
+  }
+
+  if (sock6 >= 0) {
+    FD_SET(sock6, &active_fd_set);
+  }
+
   while(1) {
-    struct sockaddr_in addr;
-    int len = sizeof(addr);
-    int fd = accept(sock, &addr, &len);
-    if (fd < 0) {
-      posixerror("accept");
+    int fd = -1;
+    int found = 0;
+    read_fd_set = active_fd_set;
+    char straddr[INET6_ADDRSTRLEN];
+
+    if (select(FD_SETSIZE, &read_fd_set, NULL, NULL, NULL) < 0) {
+      posixerror("select");
+      exit(1);
+    }
+
+    if (FD_ISSET(sock4,&read_fd_set)) {
+      struct sockaddr_in addr;
+      struct sockaddr_in addr2;
+      int len = sizeof(addr);
+
+      fd = accept(sock4, &addr, &len);
+    
+      if (fd < 0) {
+        posixerror("accept");
+        continue;
+      }
+        
+      strcpy(straddr,inet_ntoa(addr.sin_addr));
+      
+      //logmsg(LOG_INFO,"Accepted connection from %s\n", straddr);
+      found = 0;
+      struct allowed_addr *p;
+      for(p=allowed_list; p; p=p->next) {
+        addr2 = *(struct sockaddr_in *)&p->addr;
+
+        if (memcmp(&addr2.sin_addr, &addr.sin_addr, sizeof(addr.sin_addr)) == 0) {
+          //logmsg(LOG_INFO,"Found the allowed address\n");
+          found=1;
+          break;
+        }
+      }
+    }
+    else if (FD_ISSET(sock6,&read_fd_set)) {
+      struct sockaddr_in6 addr;
+      struct sockaddr_in6 addr2;
+      int len = sizeof(addr);
+      fd = accept(sock6, &addr, &len);
+      if (fd < 0) {
+        posixerror("accept");
+        continue;
+      }
+            
+      inet_ntop(AF_INET6, &addr.sin6_addr, straddr, sizeof(straddr));
+
+      //logmsg(LOG_INFO,"Accepted connection from %s\n",straddr);
+      found = 0;
+      struct allowed_addr *p;
+      for(p=allowed_list; p; p=p->next) {
+        addr2 = *(struct sockaddr_in6 *)&p->addr;
+
+        if (memcmp(&addr2.sin6_addr, &addr.sin6_addr, sizeof(addr.sin6_addr)) == 0) {
+          //logmsg(LOG_INFO,"Found the allowed address\n");
+          found=1;
+          break;
+        }
+      }
+    }
+    else {
       continue;
     }
 
-    //logmsg(LOG_INFO,"Accepted connection from %s\n", inet_ntoa(addr.sin_addr));
-    int found = 0;
-    struct allowed_addr *p;
-    for(p=allowed_list; p; p=p->next) {
-      if (memcmp(&p->addr, &addr.sin_addr, sizeof(addr.sin_addr)) == 0) {
-        //logmsg(LOG_INFO,"Found the allowed address\n");
-        found=1;
-        break;
-      }
-    }
     if (!found) {
-      logmsg(LOG_INFO,"Address %s is not on the allowed list.\n", inet_ntoa(addr.sin_addr));
+      logmsg(LOG_INFO,"Address %s is not on the allowed list.\n", straddr);
       close(fd);
       continue;
     }
